/*
 * Copyright (c) 2004 Hamish Rodda <rodda@kde.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "settingsbridge.h"

#include <qapplication.h>
#include <qdesktopwidget.h>
#include <qcheckbox.h>
#include <qregexp.h>

#include <kdebug.h>
#include <knuminput.h>
#include <kconfig.h>

#include "attributes.h"

#include <X11/Xlib.h>
#include <X11/extensions/NVCtrl.h>

extern "C" {
	#include <X11/extensions/NVCtrlLib.h>
}

#include "attributemap.h"

SettingsBridge::SettingsBridge(bool loadOnly, QObject *parent, const char *name)
	: QObject(parent, name)
	, m_loadOnly(loadOnly)
	, m_bridgeValid(false)
	, m_changed(false)
	, m_currentScreen(-1)
{
	if (!XNVCTRLQueryExtension(qt_xdisplay(), &m_eventBase, 0))
		return;
	
	int major, minor;
	if (!XNVCTRLQueryVersion(qt_xdisplay(), &major, &minor))
		return;
		
	// Minimum version == current version as of writing
	if (major < 1 || minor < 6)
		return;
	
	for (uint screen = 0; screen <= (uint)QApplication::desktop()->numScreens(); screen++) {
		if (screenValid(screen)) {
			// This bridge now valid since we have a supported screen
			m_bridgeValid = true;
			m_validScreens.append(screen);
			
			if (m_currentScreen == -1)
				m_currentScreen = screen;
			
			if (!m_loadOnly)
				XNVCtrlSelectNotify(qt_xdisplay(), screen, ATTRIBUTE_CHANGED_EVENT, true);
	
			int displays = queryAttribute(NV_CTRL_CONNECTED_DISPLAYS, 0, screen);
			QMap<QString, int> screenDisplays;
			for (int i = 0; (1 << i) <= displays && i < int(sizeof(int) * 8); ++i)
				if ((1 << i) & displays)
					screenDisplays.insert(queryString(NV_CTRL_STRING_DISPLAY_DEVICE_NAME, 1 << i, screen), 1 << i);
					
			m_displays.insert(screen, screenDisplays);
		}
	}
}

void SettingsBridge::registerWidget( QWidget * widget, uint attribute, bool reverse, QWidget * containerWidget )
{
	registerAttribute(new SimpleAttribute(this, widget, attribute, 0, reverse), containerWidget);
}

void SettingsBridge::registerWidget( QWidget * widget, uint attribute, int displayMask, bool appliesImmediately, QWidget * containerWidget )
{
	SimpleAttribute* a = new SimpleAttribute(this, widget, attribute, displayMask);
	if (appliesImmediately)
		a->setAppliesImmediately();
	registerAttribute(a, containerWidget);
}

void SettingsBridge::registerAttribute(AbstractAttribute* a, QWidget * containerWidget)
{
	if (a->isValid()) {
		m_widgets.insert(a->widget(), a);
		m_attributes.append(a);
	
		a->update();
		a->makeConnections();
		
		connect(a, SIGNAL(changed(bool)), SLOT(slotAttributeChanged(bool)));
	
	} else {
		containerWidget ? containerWidget->hide() : a->widget()->hide();
		delete a;
	}
}

void SettingsBridge::deregisterWidget( QWidget * widget )
{
	Q_ASSERT(m_widgets.contains(widget));
	AbstractAttribute* a = m_widgets[widget];
	m_widgets.remove(widget);
	m_attributes.remove(a);
	delete a;
}

bool SettingsBridge::isValid( ) const
{
	return m_bridgeValid;
}

bool SettingsBridge::screenValid( int screen ) const
{
	return screen < QApplication::desktop()->numScreens() && XNVCTRLIsNvScreen(qt_xdisplay(), screen);
}

int SettingsBridge::queryAttribute( int attribute, int displayMask, int screen ) const
{
	int ret;
	if (XNVCTRLQueryAttribute(qt_xdisplay(), screen == -1 ? currentScreen() : screen, displayMask, attribute, &ret))
		return ret;
	
	//kdDebug() << "screen " << currentScreen() << " display " << displayMask << " attr " << attribute << endl;
	
	Q_ASSERT(false);
	return -1;
}

QString SettingsBridge::queryString( int attribute, int displayMask, int screen ) const
{
	char* str;
	if (!XNVCTRLQueryStringAttribute(qt_xdisplay(), screen == -1 ? currentScreen() : screen, displayMask, attribute, &str))
		return QString::null;
	
	QString ret(str);
	delete str;
	
	return ret;
}


int SettingsBridge::currentScreen( ) const
{
	return m_currentScreen;
}

void SettingsBridge::setCurrentScreen( int screen )
{
	Q_ASSERT(screenValid(screen));
	m_currentScreen = screen;
}

void SettingsBridge::setAttribute( int attribute, int value, int displayMask, int screen )
{
	// Don't want to hear about our own changes...
	if (!m_loadOnly)
		XNVCtrlSelectNotify(qt_xdisplay(), screen == -1 ? currentScreen() : screen, ATTRIBUTE_CHANGED_EVENT, false);
	
	XNVCTRLSetAttribute(qt_xdisplay(), screen == -1 ? currentScreen() : screen, displayMask, attribute, value);
	
	if (!m_loadOnly)
		XNVCtrlSelectNotify(qt_xdisplay(), screen == -1 ? currentScreen() : screen, ATTRIBUTE_CHANGED_EVENT, true);
}

void SettingsBridge::slotAttributeChanged( bool c )
{
	if (c == m_changed)
		return;

	if (c) {
		m_changed = true;
	
	} else {
		// Need to check to see if others are changed
		for (QValueList<AbstractAttribute*>::ConstIterator it = m_attributes.begin(); it != m_attributes.end(); ++it)
			if ((*it)->isChanged())
				return;

		m_changed = false;
	}
	
	emit changed(m_changed);
}

void SettingsBridge::commitChanges( )
{
	Q_ASSERT(m_changed);

	for (QValueList<AbstractAttribute*>::ConstIterator it = m_attributes.begin(); it != m_attributes.end(); ++it)
		if ((*it)->isChanged())
			(*it)->commitChange();
	
	m_changed = false;
	emit changed(m_changed);
}

void SettingsBridge::revertChanges( )
{
	Q_ASSERT(m_changed);
	
	for (QValueList<AbstractAttribute*>::ConstIterator it = m_attributes.begin(); it != m_attributes.end(); ++it)
		if ((*it)->isChanged())
			(*it)->revertChange();
	
	m_changed = false;
	emit changed(m_changed);
}

int SettingsBridge::eventBase( ) const
{
	return m_eventBase;
}

void SettingsBridge::attributeChanged( int screen, uint displayMask, uint attribute)
{
	if (screen != currentScreen())
		return;

	for (QValueList<AbstractAttribute*>::ConstIterator it = m_attributes.begin(); it != m_attributes.end(); ++it)
		(*it)->checkUpdate(attribute, displayMask);
}

static int attributeIndex(const QString& attributeName)
{
	for (AttributeTableEntry* e = attributeTable; e->name; e++)
		if (attributeName == e->name)
			return e->attr;
	
	return -1;
}

void SettingsBridge::loadFromDisk(bool registeredOnly)
{
	KConfig config("kcmnvidiarc", true);
	
	QStringList groups = config.groupList();
	for (QStringList::ConstIterator it = groups.begin(); it != groups.end(); ++it) {
		static QRegExp screenPattern("Screen (\\d+)");
		static QRegExp attributePattern("(\\w+)\\[(.+)\\]");
		
		if (screenPattern.exactMatch(*it)) {
			int screen = screenPattern.cap(1).toInt();
			
			QMap<QString, QString> entries = config.entryMap(*it);
			
			for (QMap<QString, QString>::ConstIterator it2 = entries.begin(); it2 != entries.end(); ++it2) {
				if (attributePattern.exactMatch(it2.key())) {
					int attribute = attributeIndex(attributePattern.cap(1));
					int value = it2.data().toInt();
					int displayMask = attributePattern.cap(2).length() ? displayMaskForName(attributePattern.cap(2)) : 0;
					
					if (attribute != -1) {
						for (QMap<const QWidget*, AbstractAttribute*>::ConstIterator it3 = m_widgets.begin(); it3 != m_widgets.end(); ++it3)
							if (it3.data()->concernsAttribute(attribute)) {
								it3.data()->setAttribute(screen, attribute, value);
								goto found;
							}
						
						if (!registeredOnly) {
							SimpleAttribute* a = dynamic_cast<SimpleAttribute*>(attributeObject(attribute, displayMask));
							if (!a) {
								a = new SimpleAttribute(this, 0L, attribute, displayMaskForName(attributePattern.cap(2)));
								m_attributes.append(a);
							}
							
							a->setAttribute(screen, attribute, value);
						}
						
						found:
						// Gah, have to have a statement here
						continue;
					}
					
				} else kdWarning() << k_funcinfo << "Attribute name not matched." << endl;
			}
			
		} else kdWarning() << k_funcinfo << "Screen name not matched." << endl;
	}
}

void SettingsBridge::saveToDisk( )
{
	KConfig config("kcmnvidiarc");
	
	for (QValueList<int>::ConstIterator screen = validScreens().begin(); screen != validScreens().end(); ++screen) {
		config.setGroup(QString("Screen %1").arg(*screen));
		
		for (QValueList<AbstractAttribute*>::ConstIterator attr = m_attributes.begin(); attr != m_attributes.end(); ++attr)
			(*attr)->save(&config, *screen);
	}
}

int SettingsBridge::displayMaskForName( const QString & displayName, int screen ) const
{
	if (!m_displays.contains(screen))
		return -1;
	
	if (!m_displays[screen].contains(displayName))
		return -1;
	
	return m_displays[screen][displayName];
}

const QValueList< int > & SettingsBridge::validScreens( ) const
{
	return m_validScreens;
}

AbstractAttribute * SettingsBridge::attributeObject( int attributeIndex, int displayMask ) const
{
	for (QValueList<AbstractAttribute*>::ConstIterator it = m_attributes.begin(); it != m_attributes.end(); ++it)
		if ((*it)->concernsAttribute(attributeIndex, displayMask))
			return *it;
			
	return 0L;
}

QString SettingsBridge::attributeName( int attribute, int displayMask, int screen ) const
{
	QString mask;
	if (displayMask)
		mask = QString("[%1]").arg(displayNameForMask(displayMask, screen));
	
	uint i = 0;
	for (; attributeTable[i].name; ++i)
		if (attributeTable[i].attr == attribute)
			break;
	
	Q_ASSERT(attributeTable[i].name);
	if (!attributeTable[i].name) {
		kdWarning() << "Cannot find attribute " << attribute << endl;
		return "Error";
	}
	
	return QString("%1%2").arg(attributeTable[i].name).arg(mask);
}

QString SettingsBridge::displayNameForMask( int displayMask, int screen ) const
{
	return queryString(NV_CTRL_STRING_DISPLAY_DEVICE_NAME, displayMask, screen == -1 ? currentScreen() : screen);
}

#include "settingsbridge.moc"
